@@ -43,6 +43,7 @@ struct btrfs_fs_info *info;
 struct btrfs_root *root;
 struct btrfs_path path;
 int block_size = 0;
+uint64_t dev_size = 0;
 
 ///set useb block
 static void set_bitmap(unsigned long* bitmap, uint64_t pos, uint64_t length){
@@ -51,6 +52,10 @@ static void set_bitmap(unsigned long* bitmap, uint64_t pos, uint64_t length){
     uint64_t block_end;
 
     log_mesg(3, 0, 0, fs_opt.debug, "%s: offset: %llu size: %llu block_size: %i\n", __FILE__,  pos, length, block_size);
+    if (pos > dev_size) {
+	log_mesg(1, 0, 0, fs_opt.debug, "%s: offset(%llu) larger than device size(%llu), skip it.\n", __FILE__,  pos, dev_size);
+	return;
+    }
     pos_block = pos/block_size;
     block_end = (pos+length)/block_size;
     if ((pos+length)%block_size > 0)
@@ -307,8 +312,9 @@ extern void readbitmap(char* device, image_head image_hdr, unsigned long* bitmap
     int slot;
 
     fs_open(device);
+    dev_size = image_hdr.device_size;
     block_size  = btrfs_super_nodesize(info->super_copy);
-    
+
     set_bitmap(bitmap, BTRFS_SUPER_INFO_OFFSET, block_size);
     //check_extent_bitmap(bitmap, btrfs_root_bytenr(&info->extent_root->root_item), &block_size);
     //check_extent_bitmap(bitmap, btrfs_root_bytenr(&info->csum_root->root_item), &block_size);
@@ -391,6 +397,7 @@ extern void initial_image_hdr(char* device, image_head* image_hdr)
     image_hdr->totalblock  = (uint64_t)(image_hdr->device_size/image_hdr->block_size);
     log_mesg(0, 0, 0, fs_opt.debug, "%s: totalblock = %lli\n", __FILE__, image_hdr->totalblock);
 
+
     fs_close();
     log_mesg(0, 0, 0, fs_opt.debug, "%s: fs_close\n", __FILE__);
 }
