@@ -41,7 +41,7 @@ static void fs_open(char* device){
     int use_blocksize = 0;
     int flags;
 
-    flags = EXT2_FLAG_JOURNAL_DEV_OK | EXT2_FLAG_SOFTSUPP_FEATURES;
+    flags = EXT2_FLAG_JOURNAL_DEV_OK | EXT2_FLAG_SOFTSUPP_FEATURES | EXT2_FLAG_64BITS;
     if (use_superblock && !use_blocksize) {
 	for (use_blocksize = EXT2_MIN_BLOCK_SIZE; use_blocksize <= EXT2_MAX_BLOCK_SIZE; use_blocksize *= 2) {
 	    retval = ext2fs_open (device, flags, use_superblock, use_blocksize, unix_io_manager, &fs);
@@ -82,12 +82,12 @@ static int block_size(){
 
 /// get total block from super block
 static unsigned long long block_count(){
-    return (unsigned long long)fs->super->s_blocks_count;
+    return (unsigned long long)ext2fs_blocks_count(fs->super);
 }
 
 /// get used blocks ( total - free ) from super block
 static unsigned long long get_used_blocks(){
-    return (unsigned long long)(fs->super->s_blocks_count - fs->super->s_free_blocks_count);
+    return (unsigned long long)(ext2fs_blocks_count(fs->super) - ext2fs_free_blocks_count(fs->super));
 }
 
 /// readbitmap - cread and heck bitmap, reference dumpe2fs
@@ -134,7 +134,7 @@ extern void readbitmap(char* device, image_head image_hdr, unsigned long* bitmap
 	B_UN_INIT = 0;
 
 	if (block_bitmap) {
-	    ext2fs_get_block_bitmap_range(fs->block_map, blk_itr, block_nbytes << 3, block_bitmap);
+	    ext2fs_get_block_bitmap_range2(fs->block_map, blk_itr, block_nbytes << 3, block_bitmap);
 
 	    if (fs->super->s_feature_ro_compat & EXT4_FEATURE_RO_COMPAT_GDT_CSUM){
 #ifdef EXTFS_1_41		
@@ -153,21 +153,22 @@ extern void readbitmap(char* device, image_head image_hdr, unsigned long* bitmap
 	    for (block = 0; ((block < fs->super->s_blocks_per_group) && (current_block < (image_hdr.totalblock-1))); block++) {
 		current_block = block + blk_itr;
 
-		/// check block is used or not
-		if ((!in_use (block_bitmap, block)) || (B_UN_INIT)) {
+		if (in_use (block_bitmap, block)){
+			pc_set_bit(current_block, bitmap);
+			log_mesg(3, 0, 0, fs_opt.debug, "%s: used block %llu at group %lu\n", __FILE__, current_block, group);
+		} else {
 		    free++;
 		    gfree++;
 		    pc_clear_bit(current_block, bitmap);
 		    log_mesg(3, 0, 0, fs_opt.debug, "%s: free block %llu at group %lu init %i\n", __FILE__, current_block, group, (int)B_UN_INIT);
-		} else {
-		    pc_set_bit(current_block, bitmap);
-		    log_mesg(3, 0, 0, fs_opt.debug, "%s: used block %llu at group %lu\n", __FILE__, current_block, group);
 		}
+		
 		/// update progress
 		update_pui(&prog, current_block, current_block, 0);//keep update
 	    }
 	    blk_itr += fs->super->s_blocks_per_group;
 	}
+	log_mesg(2, 0, 0, fs_opt.debug, "%s: free bitmap (gfree = %lli, bg_blocks_count = %lli)at %lu group.\n", __FILE__, gfree, ext2fs_bg_free_blocks_count(fs, group), group);
 	/// check free blocks in group
 #ifdef EXTFS_1_41		
 	if (gfree != fs->group_desc[group].bg_free_blocks_count){	
@@ -181,11 +182,11 @@ extern void readbitmap(char* device, image_head image_hdr, unsigned long* bitmap
 	}
     }
     /// check all free blocks in partition
-    if (free != fs->super->s_free_blocks_count) {
+    if (free != ext2fs_free_blocks_count(fs->super)) {
 	if ((fs->super->s_feature_ro_compat & EXT4_FEATURE_RO_COMPAT_GDT_CSUM) && (ext4_gfree_mismatch))
 	    log_mesg(1, 0, 0, fs_opt.debug, "%s: EXT4 bitmap metadata mismatch\n", __FILE__);
 	else
-	    log_mesg(0, 1, 1, fs_opt.debug, "%s: bitmap free count err, free:%llu\n", __FILE__, free);
+	    log_mesg(0, 1, 1, fs_opt.debug, "%s: bitmap free count err, partclone get free:%llu but extfs get %llu\n", __FILE__, free, ext2fs_free_blocks_count(fs->super));
     }
 
     fs_close();
@@ -226,6 +227,12 @@ extern void initial_image_hdr(char* device, image_head* image_hdr)
     image_hdr->totalblock = (unsigned long long)block_count();
     image_hdr->usedblocks = (unsigned long long)get_used_blocks();
     image_hdr->device_size = (unsigned long long)(image_hdr->block_size * image_hdr->totalblock);
+
+    log_mesg(1, 0, 0, fs_opt.debug, "%s: extfs block_size %i\n", __FILE__, image_hdr->block_size);
+    log_mesg(1, 0, 0, fs_opt.debug, "%s: extfs total block %lli\n", __FILE__, image_hdr->totalblock);
+    log_mesg(1, 0, 0, fs_opt.debug, "%s: extfs used blocks %lli\n", __FILE__, image_hdr->usedblocks);
+    log_mesg(1, 0, 0, fs_opt.debug, "%s: extfs device size %lli\n", __FILE__, image_hdr->device_size);
+
     fs_close();
 }
 
