@@ -11,17 +11,9 @@
  * (at your option) any later version.
  */
 
-/// open device
-static void fs_open(char* device);
-
-/// close device
-static void fs_close();
-
 ///  readbitmap - read bitmap
 extern void readbitmap(char* device, image_head image_hdr, unsigned long* bitmap, int pui);
 
 /// read super block and write to image head
 extern void initial_image_hdr(char* device, image_head* image_hdr);
 
-/// return used block count
-static unsigned long long get_used_block();
